from abc import ABCMeta, abstractmethod, abstractproperty
import gettext
import locale
from typing import Iterable, Optional

from tkreform.exceptions import MessageNotFound


class Messages(dict[str, str]):
    def __matmul__(self, __o: str) -> str:
        for k in self:
            if self[k] == __o:
                return k
        raise MessageNotFound(f"Cannot find any key by value '{__o}'.")


class Linguist(metaclass=ABCMeta):
    @abstractmethod
    def __init__(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def __getitem__(self, text: str):
        raise NotImplementedError

    @abstractproperty
    def dest(self) -> str:
        raise NotImplementedError

    @dest.setter
    def dest(self, _new: str):
        raise NotImplementedError


class KVPairLinguist(Linguist):
    def __init__(
        self, lc_base: str, lc_fallback: Iterable[str], **kwargs: Messages
    ) -> None:
        self._base = lc_base
        self._messages = kwargs
        self._fallback = lc_fallback
        self.default_locale, _ = locale.getdefaultlocale()  # type: ignore
        self._dest: str = self.default_locale

    def __getitem__(self, text: str):
        key = self._messages[self._base] @ text
        return self._messages[self._dest][key]

    @property
    def dest(self) -> str:
        return self._dest

    @dest.setter
    def dest(self, _new: str):
        self._dest = _new

    @property
    def current_messages(self):
        return self._messages[self.dest]


class GettextLinguist(Linguist):
    def __init__(
        self, domain: str, dest: str, lc_fallback: Iterable[str],
        lc_path: Optional[str] = None
    ) -> None:
        self.domain = domain
        self._fallback = lc_fallback
        self._lcpath = lc_path
        self.tr = gettext.translation(
            domain, lc_path, (dest, *lc_fallback), fallback=True
        )
        self._dest = dest

    def __getitem__(self, text: str):
        return self.tr.gettext(text)

    @property
    def dest(self) -> str:
        return self._dest

    @dest.setter
    def dest(self, _new: str):
        self._dest = _new
        self.tr = gettext.translation(
            self.domain, self._lcpath, (_new, *self._fallback), fallback=True
        )



import tkinter
from typing import Any, Callable, Dict, List, Tuple, Union

from tkreform.base import Widget, Window
from tkreform.declarative import Gridder


class Group:
    def __init__(self, *content: Union["Group", Widget, Window]) -> None:
        self._contents = content

    def __getitem__(self, it: Union[int, slice]):
        return self._contents[it]


class ActionGroup(Group):
    _contents: Tuple[Union["ActionGroup", Widget, Window]]

    def __init__(self, *ct: Union["ActionGroup", Widget, Window]) -> None:
        self._calls: Dict[str, List[Callable[[tkinter.Event], Any]]] = {}
        super().__init__(*ct)

    def _setup_dict(self, seq: str):
        if seq not in self._calls:
            for co in self._contents:
                def call(event: tkinter.Event):
                    for ca in self._calls[seq]:
                        ca(event)

                co.on(seq, append=True)(call)
            self._calls[seq] = []

    def on(self, seq: str, append: bool = False):
        self._setup_dict(seq)

        def __wrapper(func: Callable[[tkinter.Event], Any]):
            if append:
                self._calls[seq].append(func)
            else:
                self._calls[seq] = [func]
            return func
        return __wrapper


class GriddingGroup(Group):
    _contents: Tuple[Widget]

    def grid(self, columnspan: int, **kwargs):
        for idx, wi in enumerate(self._contents):
            wi.grid(
                column=idx // columnspan, columnspan=columnspan,
                row=idx % columnspan, **kwargs
            )

    def __mul__(self, other: Gridder):
        if other.columnspan is None:
            raise ValueError(
                "'GriddingGroup' requires 'Gridder' to have columnspan."
            )
        self.grid(
            columnspan=other.columnspan, rowspan=other.rowspan,
            ipadx=other.ipadx, ipady=other.ipady, padx=other.padx,
            pady=other.pady, sticky=other.sticky
        )
        return self



from typing import Tuple, Union

DETAIL = 1
TYPE = 2
MODIFIER = 3
NMODIFIER = 4

KEY_LITERAL_SEP = "+"


_negative_event = {
    "Key": "KeyRelease",
    "KeyRelease": "Key",
    "KeyPress": "KeyRelease",
    "Button": "ButtonRelease",
    "ButtonRelease": "Button",
    "ButtonPress": "ButtonRelease",
    "Enter": "Leave",
    "Leave": "Enter",
    "FocusIn": "FocusOut",
    "FocusOut": "FocusIn"
}


class Event:
    def __init__(self, *eventdata: Tuple[str, str, int]) -> None:
        _eventdata = set(eventdata)
        self._ed = sorted(_eventdata, key=lambda x: (-x[2], x[1]))

    @property
    def literal(self):
        return KEY_LITERAL_SEP.join(x[0] for x in self._ed if x[0])

    @property
    def event(self):
        return f"<{'-'.join(x[1] for x in self._ed if x[1])}>"

    def __str__(self) -> str:
        return self.event

    def __repr__(self) -> str:
        return f"Event({self.event})"

    def __hash__(self):
        return hash(self.event)

    def __add__(self, other: Union["Event", int, str]):
        if not isinstance(other, Event):
            _other = str(other)
            o = type(self)((_other, _other, DETAIL))
        else:
            o = other
        return type(self)(*self._ed, *o._ed)

    def __sub__(self, other: Union["Event", int, str]):
        return self + other

    def __and__(self, other: Union["Event", int, str]):
        return self + other

    def __or__(self, other: Union["Event", int, str]):
        return self + other

    def __neg__(self):
        return type(self)(
            *tuple(
                (name, _negative_event[event], ord)
                if event in _negative_event else
                (name, event, ord) for name, event, ord in self._ed
            )
        )


KEY = Event(("", "Key", TYPE))
CTRL = Event(("Ctrl", "Control", MODIFIER))
ALT = Event(("Alt", "Alt", MODIFIER))
SHIFT = Event(("Shift", "Shift", MODIFIER))
CAPS = Event(("CapsLock", "Lock", MODIFIER))


def FN(n: int):
    return Event((f"F{n}", f"F{n}", TYPE))


X2 = Event(("", "Double", NMODIFIER))
X3 = Event(("", "Triple", NMODIFIER))
X4 = Event(("", "Quadruple", NMODIFIER))

LMB = Event(("Left Mouse Button", "Button", TYPE), ("", "1", DETAIL))
RMB = Event(("Right Mouse Button", "Button", TYPE), ("", "2", DETAIL))
CMB = Event(("Middle Mouse Button", "Button", TYPE), ("", "3", DETAIL))



from tkinter import Menu
from tkreform.exceptions import MenuNotBinded


class MenuItem:
    def __init__(self, type: str, **kwargs):
        self.type = type
        self.data = kwargs
        self.base = None

    def bind_menu(self, m: Menu, **kwargs):
        # self.base = Menu(m, **kwargs)
        self.base = m

    def _add_item(self, type: str, **kwargs):
        if self.base is None:
            raise MenuNotBinded
        self.base.add(type, **kwargs)

    def add_item(self, it: "MenuItem"):
        if self.base is None:
            raise MenuNotBinded
        it.bind_menu(self.base)
        self._add_item(it.type, **it.data)


class MenuCascade(MenuItem):
    def __init__(self, **kwargs):
        super().__init__("cascade", **kwargs)


class MenuCheckbutton(MenuItem):
    def __init__(self, **kwargs):
        super().__init__("checkbutton", **kwargs)


class MenuCommand(MenuItem):
    def __init__(self, **kwargs):
        super().__init__("command", **kwargs)


class MenuRadioButton(MenuItem):
    def __init__(self, **kwargs):
        super().__init__("radiobutton", **kwargs)


class MenuSeparator(MenuItem):
    def __init__(self, **kwargs):
        super().__init__("separator", **kwargs)



"""
Reformed tkinter coding tool.

TkReform provides a wrapping on standard Tkinter. Window wraps Toplevel
widgets, and Widget wraps other widgets.

Properties of the widgets can be specified with keyword arguments or
attributes.

Example (Hello, World):
>>> import tkinter as tk
>>> import tkreform
>>> window = tkreform.Window(tk.Tk())
>>> frame = window.add_widget(tk.Frame, relief="ridge", borderwidth=2)
>>> frame.pack(fill="both", expand=True)
>>> label = frame.add_widget(tk.Label)
>>> label.text = "Hello, World"
>>> label.pack(fill="x", expand=True)
>>> button = frame.add_widget(tk.Button, text="Exit")
>>> button.callback(window.destroy)
>>> button.pack(side="bottom")
>>> window.loop()
"""

from tkreform import base, declarative, groups
from tkreform.base import dec, Widget, Window
from tkreform.declarative import Gridder, Packer, Placer

__all__ = [
    "base", "dec", "declarative", "groups", "Widget", "Window", "Gridder",
    "Packer", "Placer"
]
